.-panel-launch-button
{
    padding: 2px;
    margin: 0px;
    border-radius: 2px;
}
.-panel-launch-button:hover,
.-panel-launch-button:hover,
.-panel-launch-button.highlight,
.-panel-launch-button:active:hover,
.-panel-launch-button-selected
{
    -gtk-icon-effect: highlight;
}

.-panel-launch-button:selected
{
    outline-offset: -2px;
}
